package mainForm;

import java.io.IOException;
import java.util.ArrayList;

public class Generator {
    private int bpm;
    private int barCount;
    private int keyShift;
    private boolean moll;

    private int[] getTones() {
        if (moll) {
            return GeneratorUtil.getMinorTones();
        } else {
            return GeneratorUtil.getMajorTones();
        }
    }

    Generator(int bpm, int barCount, int keyShift, boolean moll) {
        this.bpm = bpm;
        this.barCount = barCount;
        this.keyShift = keyShift;
        this.moll = moll;
    }

    public void generate(String path) throws IOException {
        MIDIFile midiFile = new MIDIFile();

        MIDITrack tempoTrack = new MIDITrack();

        tempoTrack.getEventsList().add(MIDIEvent.timeSignature(0, 4, 4));
        tempoTrack.getEventsList().add(MIDIEvent.keySignature(0, keyShift, !moll));
        tempoTrack.getEventsList().add(MIDIEvent.tempoSignature(0, bpm));
        tempoTrack.getEventsList().add(MIDIEvent.trackFooter(1));
        midiFile.getTrackList().add(tempoTrack);

        MIDITrack musicTrack = new MIDITrack();
        musicTrack.getEventsList().add(MIDIEvent.programChange(0, 0,1));

        Score song = new Score();

        int periodCount = barCount / 8;
        for (int i = 0; i < periodCount; i++) {
            song.add(GeneratorUtil.getRandomPeriod(getTones(), i == periodCount -1));
        }

        song.keyShift(keyShift);

        writeScoreToMIDITrack(song, musicTrack);
        musicTrack.getEventsList().add(MIDIEvent.trackFooter(0));
        midiFile.getTrackList().add(musicTrack);
        midiFile.generateMIDIFile(path);
    }

    private void writeScoreToMIDITrack(Score score, MIDITrack midiTrack){
        class ToneInArray{
            private int tone;
            private double shift;
            private boolean on;

            private ToneInArray(int tone, double shift, boolean on) {
                this.tone = tone;
                this.shift = shift;
                this.on = on;
            }
        }

        ArrayList<ToneInArray> tonesArray = new ArrayList<>();

        for (NoteList part : score) {
            double counter = 0;

            for (Note note : part) {
                int index = tonesArray.size();
                for (int j = 0; j < tonesArray.size(); j++) {
                    if (tonesArray.get(j).shift > counter){
                        index = j;
                        break;
                    }
                }

                if (note.getTone() != -1) {
                    tonesArray.add(index, new ToneInArray(note.getTone(), counter, true));
                }
                counter = counter + note.getDuration();
                index = tonesArray.size();
                for (int j = 0; j < tonesArray.size(); j++) {
                    if (tonesArray.get(j).shift > counter){
                        index = j;
                        break;
                    }
                }
                if (note.getTone() != -1) {
                    tonesArray.add(index, new ToneInArray(note.getTone(), counter, false));
                }
            }
        }


        for (int i = 0; i < tonesArray.size(); i++) {
            double delta;

            if (i == 0){
                delta = 0;
            } else {
                delta = (tonesArray.get(i).shift - tonesArray.get(i - 1).shift) * 4 * MIDIFile.getPPQ();
                if (delta < 0){
                    delta = 0;
                }
            }

            if (tonesArray.get(i).on){
                midiTrack.getEventsList().add(
                        MIDIEvent.noteOn((int)Math.round(delta), tonesArray.get(i).tone, 100/*127*/));
            } else {
                midiTrack.getEventsList().add(
                        MIDIEvent.noteOff((int)Math.round(delta), tonesArray.get(i).tone));
            }
        }
    }
}
