package mainForm;

import java.util.*;
import java.util.stream.IntStream;

public class GeneratorUtil {
    private static final double WHOLE = 1;
    private static final double HALF = 1d / 2d;
    private static final double QUARTER = 1d / 4d;
    private static final double EIGHTH = 1d / 8d;
    private static final double SIXTEENTH = 1d / 16d;

    private static final int TONIC_DEGREE = 0;
    private static final int SUBDOMINANT_DEGREE = 3;
    private static final int DOMINANT_DEGREE = 4;

    private static int[] majorTones = {0, 2, 4, 5, 7, 9, 11};
    private static int[] minorTones = {0, 2, 3, 5, 7, 8, 10};

    public static double getWHOLE() {
        return WHOLE;
    }
    public static double getHALF() {
        return HALF;
    }
    public static double getQUARTER() {
        return QUARTER;
    }
    public static double getEIGHTH() {
        return EIGHTH;
    }
    public static double getSIXTEENTH() {
        return SIXTEENTH;
    }

    public static int[] getMajorTones() { return majorTones; }
    public static int[] getMinorTones() { return minorTones; }

    private static ArrayList<Integer> getScale(int lowestTone, int highestTone, int[] tones){
        if (lowestTone > highestTone){
            throw new IllegalArgumentException("lowestTone is higher than highestTone!");
        }

        if (lowestTone < 0){
            throw new IllegalArgumentException("lowestTone must be 0 or higher!");
        }
        if (highestTone > 127){
            throw new IllegalArgumentException("highestTone must be 127 or smaller!");
        }

        ArrayList<Integer> arrayList = new ArrayList<>();
        for (int i = lowestTone; i < highestTone +1 ; i++) {
            int modulo = i % 12;
            if (IntStream.of(tones).anyMatch(x -> x == modulo)) {
                arrayList.add(i);
            }
        }
        return arrayList;
    }

    public static Score getRandomPeriod(int [] tones, boolean lastNoteIsTonic){
        int lowestMelodyTone = 60;
        int highestMelodyTone = 84;
        int lowestAccompanimentTone = 48;

        ArrayList<Motif> motivicStructure = getMotivicStructure();
        ArrayList<ArrayList<Double>> rhythmicStructure = getRhythmicStructure(motivicStructure);
        int [] harmonicStructure = getHarmonicStructure();
        NoteList melody = getRandomMelody(tones,
                                          motivicStructure,
                                          harmonicStructure,
                                          rhythmicStructure,
                                          lowestMelodyTone,
                                          highestMelodyTone,
                                          lastNoteIsTonic);

        Score result = new Score();
        result.add(melody);

        result.add(new NoteList());
        result.add(new NoteList());
        result.add(new NoteList());

        for (int harmonicDegree : harmonicStructure) {
            int[] chord = getChord(tones, harmonicDegree);
            for (int j = 1; j <= 3; j++) {
                Note n = new Note(chord[j -1] + lowestAccompanimentTone, 1);
                result.get(result.size() -j).add(n);
            }
        }

        return result;
    }

    private static NoteList getRandomMelody(int[] tones,
                                            ArrayList<Motif> motivicStructure,
                                            int[] harmonicStructure,
                                            ArrayList<ArrayList<Double>> rhythmicStructure,
                                            int lowestTone,
                                            int highestTone,
                                            boolean lastNoteIsTonic){
        NoteList result = new NoteList();
        ArrayList<NoteList> barList = new ArrayList<>();
        int lastTone = -1;

        for (Motif motif : motivicStructure) {
            NoteList bar;

            for (int j = 0; j < motif.getBarCount(); j++) {
                bar = getRandomBarMelody(tones,
                        rhythmicStructure.get(barList.size()),
                        harmonicStructure[barList.size()],
                        lowestTone,
                        highestTone,
                        lastTone,
                        lastNoteIsTonic && barList.size() == rhythmicStructure.size() -1);
                barList.add(bar);
                result.addNotes(bar);
                lastTone = bar.get(bar.size() -1).getTone();
            }
        }
        return result;
    }

    private static NoteList getRandomBarMelody(int [] tones,
                                               ArrayList<Double> barRhythm,
                                               int harmonicDegree,
                                               int lowestTone,
                                               int highestTone,
                                               int lastTone,
                                               boolean lastNoteIsTonic){
        Random rand = new Random();
        NoteList result = new NoteList();
        double lengthCounter = 0;
        for (int i = 0; i < barRhythm.size(); i++) {
            double duration = barRhythm.get(i);
            int tone;
            ArrayList<Integer> possibleTones;
            ArrayList<Integer> allTones = getScale(lowestTone, highestTone, tones);
            ArrayList<Integer> chordTones = getScale(lowestTone, highestTone, getChord(tones, harmonicDegree));
            ArrayList<Integer> tonics = getScale(lowestTone, highestTone, new int[] {tones[0]});

            if (i == barRhythm.size() -1){
                possibleTones = new ArrayList<>(tonics);
            } else if ((Math.round(lengthCounter * 1000) % 4000) == 0 || duration > QUARTER + 0.05){
                possibleTones = new ArrayList<>(chordTones);
            } else {
                possibleTones = new ArrayList<>(allTones);
            }

            if (possibleTones.contains(lastTone)){
                possibleTones.remove(possibleTones.indexOf(lastTone));
            }

            if (lastNoteIsTonic && i == barRhythm.size() -2){
                for (int tonic : tonics) {
                    if (possibleTones.contains(tonic)){
                        possibleTones.remove(possibleTones.indexOf(tonic));
                    }
                }
            }

            if (lastTone == -1){
                tone = possibleTones.get(Math.abs(rand.nextInt() % possibleTones.size()));
            } else {
                boolean ascending = (lastTone <= possibleTones.get(0))
                                 || ((lastTone < possibleTones.get(possibleTones.size() -1)) && rand.nextBoolean());

                boolean allowStep = false;
                for (int possibleTone : possibleTones) {
                    int diff = allTones.indexOf(lastTone) - allTones.indexOf(possibleTone);
                    if (ascending){
                        diff = - diff;
                    }

                    if (diff == 1){
                        allowStep = true;
                        break;
                    }
                }

                boolean allowLeap = false;
                for (int chordTone : chordTones) {
                    int diff = allTones.indexOf(lastTone) - allTones.indexOf(chordTone);
                    if (ascending){
                        diff = - diff;
                    }

                    boolean a = !lastNoteIsTonic || i != barRhythm.size() -1 || tonics.contains(chordTone);
                    boolean b = !lastNoteIsTonic || i != barRhythm.size() -2 || !tonics.contains(chordTone);

                    if (diff > 1 && a && b){
                        allowLeap = true;
                        break;
                    }
                }

                if (!allowLeap || (allowStep && rand.nextBoolean())){
                    if (ascending){
                        tone = allTones.get(allTones.indexOf(lastTone) +1);
                    } else {
                        tone = allTones.get(allTones.indexOf(lastTone) -1);
                    }
                } else {
                    int leap = (int)Math.floor(Math.abs(rand.nextGaussian())) +2;
                    int minLeap = -1;
                    int maxLeap = -1;
                    possibleTones = new ArrayList<>(chordTones);

                    if (lastNoteIsTonic){
                        if (i == barRhythm.size() -1){
                            possibleTones = new ArrayList<>(tonics);

                        } else if (i == barRhythm.size() -2) {

                            for (int tonic: tonics) {
                                if (possibleTones.contains(tonic)) {
                                    possibleTones.remove(possibleTones.indexOf(tonic));
                                }
                            }
                        }
                    }

                    for (int j = possibleTones.size() -1; j >= 0; j--) {
                        int diff = allTones.indexOf(lastTone) - allTones.indexOf(possibleTones.get(j));
                        if (ascending){
                            diff = - diff;
                        }

                        if (diff <= 1){
                            possibleTones.remove(j);
                        } else {
                            if (minLeap == -1 || diff < minLeap) {
                                minLeap = diff;
                            }

                            if (maxLeap == -1 || diff > maxLeap) {
                                maxLeap = diff;
                            }
                        }
                    }

                    if (leap < minLeap){
                        leap = minLeap;
                    }

                    if (leap > maxLeap){
                        leap = maxLeap;
                    }

                    tone = -1;
                    for (int possibleTone : possibleTones) {
                        int diff = allTones.indexOf(lastTone) - allTones.indexOf(possibleTone);
                        if (ascending){
                            diff = - diff;
                        }

                        if (diff <= leap){
                            tone = possibleTone;
                        }
                    }
                }
            }

            result.add(new Note(tone, duration));
            lastTone = tone;
            lengthCounter = lengthCounter + duration;
        }

        return result;
    }

    private static int [] getHarmonicStructure(){
        int [] result = new int[8];
        Random rand = new Random();

        for (int i = 7; i >=0; i--) {
            if (i == 7){
                result[i] = TONIC_DEGREE;
            } else if (i == 0){
                if (result[i +1] == DOMINANT_DEGREE){
                    result[i] = TONIC_DEGREE;
                } else if (result[i +1] == TONIC_DEGREE){
                    result[i] = DOMINANT_DEGREE;
                } else if (rand.nextBoolean()){
                    result[i] = TONIC_DEGREE;
                } else {
                    result[i] = DOMINANT_DEGREE;
                }
            } else {
                if (rand.nextBoolean()){
                    switch (result[i +1]){
                        case TONIC_DEGREE:
                            result[i] = SUBDOMINANT_DEGREE;
                            break;
                        case SUBDOMINANT_DEGREE:
                            result[i] = DOMINANT_DEGREE;
                            break;
                        case DOMINANT_DEGREE:
                            result[i] = TONIC_DEGREE;
                            break;
                    }
                } else {
                    switch (result[i +1]){
                        case TONIC_DEGREE:
                            result[i] = DOMINANT_DEGREE;
                            break;
                        case SUBDOMINANT_DEGREE:
                            result[i] = TONIC_DEGREE;
                            break;
                        case DOMINANT_DEGREE:
                            result[i] = SUBDOMINANT_DEGREE;
                            break;
                    }
                }
            }
        }

        return result;
    }

    private static ArrayList<ArrayList<Double>> getRhythmicStructure(ArrayList<Motif> motivicStructure){
        ArrayList<ArrayList<Double>> result = new ArrayList<>();
        Random rand = new Random();

        for (int i = 0; i < motivicStructure.size(); i++) {
            if (motivicStructure.get(i).getRelatedMotif() == null){
                for (int j = 0; j < motivicStructure.get(i).getBarCount(); j++) {
                    result.add(getRandomBarRhythm());
                }
            } else {
                int index = 0;
                for (Motif motif : motivicStructure) {
                    if (motif == motivicStructure.get(i).getRelatedMotif()){
                        int last = result.size() -1;
                        for (int j = index; j < index +motivicStructure.get(i).getBarCount(); j++) {
                            ArrayList<Double> bar = new ArrayList<>();
                            if (motivicStructure.get(i).isRetrograde()){
                                for (int k = result.get(j).size() -1; k >= 0; k--) {
                                    bar.add(result.get(j).get(k));
                                }
                                result.add(last, bar);
                            } else {
                                bar.addAll(result.get(j));
                                result.add(bar);
                            }
                        }
                        break;
                    }
                    index = index + motif.getBarCount();
                }
            }
        }

        if (rand.nextBoolean()){
            result.get(7).clear();
            result.get(7).add(WHOLE);
        } else {
            double space = 0;
            double last;
            for (int i = result.get(7).size() - 1; i >= 0; i--) {
                last = result.get(7).get(result.get(7).size() - 1);
                if (space + last < HALF + 0.005) {
                    space = space + last;
                    result.get(7).remove(result.get(7).size() - 1);
                } else {
                    break;
                }
            }

            last = result.get(7).get(result.get(7).size() - 1);
            result.get(7).set(result.get(7).size() - 1, last - (HALF - space));
            result.get(7).add(HALF);
        }

        return result;
    }

    private static ArrayList<Double> getRandomBarRhythm(){
        ArrayList<Double> bar = new ArrayList<>();
        Random rand = new Random();
        double length = 1;

        while (Math.round(length * 1000) > 0){
            int optionCount;
            if (length > HALF -0.005){
                optionCount = 3;
            } else if (length > QUARTER -0.005){
                optionCount = 2;
            } else {
                optionCount = 1;
            }

            double noteLength = 0;
            switch (Math.abs(rand.nextInt()) % optionCount){
                case 0:
                    noteLength = EIGHTH;
                    break;
                case 1:
                    noteLength = QUARTER;
                    break;
                case 2:
                    noteLength = HALF;
                    break;
            }

            bar.add(noteLength);
            length = length - noteLength;
        }

        return bar;
    }

    private static ArrayList<Motif> getMotivicStructure(){
        ArrayList<Motif> result = new ArrayList<>();
        ArrayList<Motif> possibleRelatedMotifs = new ArrayList<>();
        Random rand = new Random();
        int length = 0;

        while (length < 8){
            Motif relatedMotif = null;
            if (rand.nextBoolean() && possibleRelatedMotifs.size() > 0){
                int index = Math.abs(rand.nextInt()) % possibleRelatedMotifs.size();
                relatedMotif = possibleRelatedMotifs.remove(index);
            }

            int barCount;
            if (length == 7 || rand.nextBoolean() || (relatedMotif != null && relatedMotif.getBarCount() == 1)){
                barCount = 1;
            } else {
                barCount = 2;
            }

            Motif motif = new Motif(relatedMotif, rand.nextBoolean(), barCount);
            result.add(motif);
            if (relatedMotif == null){
                possibleRelatedMotifs.add(motif);
            }

            length = length + barCount;
        }

        return result;
    }

    private static int[] getChord(int[] tones, int index){
        if (index < 0 || index >= tones.length)
            throw new IllegalArgumentException("Index out of bounds!");

        int[] result = new int[3];
        for (int i = 0; i < result.length; i++) {
            result[i] = tones[index];
            index = index + 2;
            if (index >= tones.length) {
                index = index - tones.length;
            }
        }
        return result;
    }

    private GeneratorUtil(){}
}
