package mainForm;

public class MIDIEvent {
    private int deltaPPQ;

    private int[] data;

    private static final int MIDI_CLOCKS_PER_QUARTER = 24;
    private static final int THIRTY_SECONDS_PER_QUARTER = 8;

    private MIDIEvent(int deltaPPQ, int[] data) {
        this.deltaPPQ = deltaPPQ;
        this.data = data;
    }

    public static MIDIEvent timeSignature(int deltaPPQ, int numerator, int denominator){
        double denominatorLog = Math.log(denominator)/Math.log(2);
        int[] buffer = {0xFF, 0x58, 0x4, numerator, (int)denominatorLog,
                        MIDI_CLOCKS_PER_QUARTER, THIRTY_SECONDS_PER_QUARTER};
        return new MIDIEvent(deltaPPQ, buffer);
    }

    public static MIDIEvent tempoSignature(int deltaPPQ, int bpm){
        double microseconds = 60000000 / bpm;
        byte[] a = MIDIUtil.longToBytes(Math.round(microseconds), 3);
        int[] buffer = {0xFF, 0x51, 0x3, 0, 0, 0};

        for (int i = 0; i < 3; i++) {
            buffer[i +3] = a[i];
        }

        return new MIDIEvent(deltaPPQ, buffer);
    }

    public static MIDIEvent keySignature(int deltaPPQ, int sharpsOrFlats, boolean major){
        int a;
        if (major){a = 0;} else {a = 1;}
        return new MIDIEvent(deltaPPQ, new int[]{0xFF, 0x59, 0x2, sharpsOrFlats, a});
    }

    public static MIDIEvent trackFooter(int deltaPPQ){
        return new MIDIEvent(deltaPPQ, new int[] {0xFF, 0x2F, 0x0});
    }

    public static MIDIEvent programChange(int deltaPPQ, int programNumber, int patchNumber){
        return new MIDIEvent(deltaPPQ, new int[] {0xC0 + programNumber, patchNumber});
    }

    public static MIDIEvent noteOn(int deltaPPQ, int noteNumber, int velocity){
        return new MIDIEvent(deltaPPQ, new int[] {0x90, noteNumber, velocity});
    }

    public static MIDIEvent noteOff(int deltaPPQ, int noteNumber){
        return new MIDIEvent(deltaPPQ, new int[] {0x80, noteNumber, 0});
    }

    public int byteCount(){
        return MIDIUtil.byteCount(MIDIUtil.encodeVLQ(deltaPPQ)) + data.length;
    }

    public byte[] getBytes(){
        byte[] buffer = new byte[byteCount()];
        long deltaVLQ = MIDIUtil.encodeVLQ(deltaPPQ);
        byte[] deltaBytes = MIDIUtil.longToBytes(deltaVLQ, MIDIUtil.byteCount(deltaVLQ));

        for (int i = 0; i < buffer.length; i++){
            if (i < deltaBytes.length){
                buffer[i] = deltaBytes[i];
            } else {
                buffer[i] = (byte) data[i - deltaBytes.length];
            }
        }

        return buffer;
    }
}
