package mainForm;

import javafx.collections.FXCollections;
import javafx.collections.ObservableList;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Random;

public class Model {
    private static final int MIN_BPM = 1;
    private static final int MAX_BPM = 999;
    private static final int MIN_BAR_COUNT = 8;
    private static final int MAX_BAR_COUNT = 992;

    public static int getMIN_BPM() { return MIN_BPM; }
    public static int getMAX_BPM() { return MAX_BPM; }
    public static int getMIN_BAR_COUNT() { return MIN_BAR_COUNT; }
    public static int getMAX_BAR_COUNT() { return MAX_BAR_COUNT; }

    private int keyIndex;
    private int modeIndex;
    private int bpm;
    private int barCount;
    private String targetFilePath;
    private Boolean closeApp;
    private Boolean openFolder;
    private ObservableList<String> keys;
    private ObservableList<String> modes;
    private ObservableList<String> filePaths;

    public int getKeyIndex() { return keyIndex; }
    public int getModeIndex() { return modeIndex; }
    public int getBpm() {return bpm;}
    public int getBarCount() {return barCount;}
    public String getTargetFilePath() {return targetFilePath;}
    public Boolean getCloseApp() {return closeApp;}
    public Boolean getOpenFolder() {return openFolder;}
    public ObservableList<String> getKeys() { return keys; }
    public ObservableList<String> getModes() { return modes; }
    public ObservableList<String> getFilePaths() { return filePaths; }

    public void setBarCount(int barCount) {this.barCount = (barCount / 8) * 8;}
    public void setTargetFilePath(String targetFilePath) {this.targetFilePath = targetFilePath;}
    public void setCloseApp(Boolean closeApp) {this.closeApp = closeApp;}
    public void setBpm(int bpm) {this.bpm = bpm;}
    public void setOpenFolder(Boolean openFolder) {this.openFolder = openFolder;}


    public String targetDistinctFilePath(String directory){
        if (!Files.isDirectory(Paths.get(directory)))
            throw new IllegalArgumentException("path is not a directory");

        int i = 1;
        String target = directory;

        if (!target.substring(target.length() -1).equals("\\") && !target.substring(target.length() -1).equals("/"))
            target = target + "\\";

        target = target + "track";

        while (Files.exists(Paths.get(target + Integer.toString(i) + ".midi"))) {
            i++;
        }

        return target + Integer.toString(i) + ".midi";
    }

    private Model() {
        Random rand = new Random();

        this.bpm = 100 + (int)Math.round(rand.nextGaussian() * 15);
        if (this.bpm < MIN_BPM){
            this.bpm = MIN_BPM;
        } else if (this.bpm > MAX_BPM){
            this.bpm = MAX_BPM;
        }

        this.barCount = 24;
        this.keys = FXCollections.observableArrayList("7♭","6♭","5♭","4♭","3♭","2♭","♭","-",
                                                      "#","2#","3#","4#","5#","6#","7#");
        this.modes = FXCollections.observableArrayList("dur","moll");

        this.modeIndex = Math.abs(rand.nextInt()) % 2;

        this.keyIndex = 7 + (int)Math.round(rand.nextGaussian() * 2);
        if (this.keyIndex < 0){
            this.keyIndex = 0;
        } else if (this.keyIndex >= this.keys.size()){
            this.keyIndex = this.keys.size() -1;
        }

        this.openFolder = true;
        this.closeApp = true;

        this.filePaths = FXCollections.observableArrayList();

        Path path = Paths.get(System.getProperty("user.home") + "\\Music");
        if (Files.exists(path))
            this.filePaths.add(path.toString());

        path = Paths.get(System.getProperty("user.home") + "\\Documents");
        if (Files.exists(path))
            this.filePaths.add(path.toString());

        this.filePaths.add(System.getProperty("user.dir"));

        this.targetFilePath = targetDistinctFilePath(this.filePaths.get(0));
    }

    private static Model instance;

    public static Model getInstance(){
        if (instance == null){
            instance = new Model();
        }

        return instance;
    }
}
